/* Adam Rogalewicz
 * INCLUDER project
 * distributed under GNU GPL v3 */


#include "main.h"
#include <iostream>
#include<assert.h>
#include<string.h>
//#include <sstream>

// Global variables
int DWA_VERBOSE=0;
int DWA_PRINT_CE=0;
int DWA_LOCAL_REF=1;
msat_type DWA_VARIABLE_TYPE;
TRANS_SETTING apply_trans_settings;


void print_path_from_root(product_state *s){
	if (s->prev!=NULL) { 
		print_path_from_root(s->prev);
		std::cout << "--"<< s->symbol<<"--> ";
	}
	std::cout << *s;
}

static void print_model(msat_env env)
{
    /* we use a model iterator to retrieve the model values for all the
     * variables, and the necessary function instantiations */
    msat_model_iterator iter = msat_create_model_iterator(env);
    assert(!MSAT_ERROR_MODEL_ITERATOR(iter));

    std::cout << "-----------\nCounterexample model:\n";
    while (msat_model_iterator_has_next(iter)) {
        msat_term t, v;
        char *s;
        msat_model_iterator_next(iter, &t, &v);
        s = msat_term_repr(t);
        assert(s);
	std::cout << s << " = ";
        msat_free(s);
        s = msat_term_repr(v);
        assert(s);
	std::cout << s <<", ";
        msat_free(s);
    }
    std::cout <<"\n-------------\n";
    msat_destroy_model_iterator(iter);
}

int check_sat(msat_env env, msat_term t) {
	int res;
	int group_a;
	msat_result result;
	res=msat_reset_env(env);
	assert(res == 0);
	group_a = msat_create_itp_group(env);
	res = msat_set_itp_group(env, group_a);
	assert(res == 0);
	res = msat_assert_formula(env,t); 
  	assert(res == 0);
	result=msat_solve(env);
	res=msat_reset_env(env);
	assert(res == 0);
	if (result==MSAT_SAT) return 1;
	return 0;
}

int check_term_implication(msat_env env, msat_term t1, msat_term t2,int no_reset){
// no_reset = 1 then msat env is not reseted after the check
	int res;
	msat_term tmp;
	msat_result result;
	res=msat_reset_env(env);
	assert(res==0);
	int group_a;
	group_a = msat_create_itp_group(env);
	res = msat_set_itp_group(env, group_a);
	assert(res == 0);
	res = msat_assert_formula(env,t1); 
 	assert(res == 0);
	tmp=msat_make_not(env,t2);
	res = msat_assert_formula(env,tmp); 
	assert(res == 0);
	result=msat_solve(env);
	if (!no_reset) res=msat_reset_env(env); // reset the enviroment
	assert(res==0);
	if (result==MSAT_SAT)  return 0; //implication does not hold
	return 1; //implication holds
}

int check_term_equivalence(msat_env env, msat_term t1, msat_term t2){
	int res;
	msat_term tmp;
	tmp=msat_make_not(env,t1);
	tmp=msat_make_iff(env,tmp,t2);
	if (check_sat(env, tmp)) return 0;
	return 1;
}

std::set<std::string> str_set_union(std::set<std::string> v1, std::set<std::string> v2) {
/* make union of two set of strings  */
	std::set<std::string> res;
	for(std::set<std::string>::iterator i=v1.begin();i!=v1.end();++i) {
		res.insert(*i);
	}
	for(std::set<std::string>::iterator i=v2.begin();i!=v2.end();++i) {
		res.insert(*i);
	}
	return res;

}


msat_term create_equalities(msat_env env, product_automaton aut, int aut_no, std::string symbol) {
	/* for each varable x defined in automaton "aut_no" 
	 * and not defined in some other automaton, where symbol is part of the alphabet
	 * vars create formula x0=x1.  
	 * Return conjunction of these formula 
	 */
	msat_decl d1,d2;
	std::string str;
	msat_term f1,f2,result;
	result=msat_make_true(env);
	for(std::set<std::string>::iterator i=aut.components[aut_no].vars.begin();i!=aut.components[aut_no].vars.end();++i) {
		// check, whether the var "*i" is not used in some of the automaton, where "symbol" is part of the alphabet
		int find=0;
		for(int j=0; j<aut.components.size(); j++) {
			if (aut.components[j].alphabet.find(symbol)!=aut.components[j].alphabet.end())
				if (aut.components[j].vars.find(*i)!=aut.components[j].vars.end()) {
					find=1;
					break;
				}
		}
		if (find) continue;
		str=*i+"0";
		d1=msat_find_decl(env,str.c_str());
		assert(!MSAT_ERROR_DECL(d1));
		str=*i+"1";
		d2=msat_find_decl(env,str.c_str());
		assert(!MSAT_ERROR_DECL(d2));
		f1=msat_make_constant(env,d1);
		f2=msat_make_constant(env,d2);
		f1=msat_make_equal(env,f1,f2);
		result=msat_make_and(env,result,f1);
	}
	return result;
}


std::vector<product_state *> path_from_root(product_state *s){
/* get a path from root th the current product state */
	std::vector<product_state *> path;
	path.push_back(s);
	while (s->prev!=NULL){
		s=s->prev;
		path.push_back(s);		
	}
	return path;
}


msat_term get_label(int from, std::string symbol,int to,automaton aut){
/* get an label of the automata transition "from --- symbol ---> to" */
	std::vector<a_rhs>::iterator it;
	a_lhs tmp;
	tmp.src=from; tmp.lab=symbol;
	for(it=aut.trans[tmp].begin();it!=aut.trans[tmp].end();++it) {
		a_rhs tmp2;
		tmp2=*it;
		if (tmp2.dest==to)
			return tmp2.formula;
	}
	assert(0);
}

void AddMissingVar(msat_env env, std::string v1,msat_type type) {
/* add variable into enviroment if it does not exists */
	msat_decl d;
	d=msat_find_decl(env,v1.c_str());
	if(MSAT_ERROR_DECL(d)) {
		/* does not exists -> create */
		d = msat_declare_function(env, v1.c_str(), type);
        	assert(!MSAT_ERROR_DECL(d));
	} 
}

msat_term *get_variables_to_quantify_out(msat_env env, int from, int to, std::set<std::string> vars) {
	msat_term *list;
	//list=(msat_term *) malloc(sizeof(msat_term)*(to-from+1)*vars.size());
	list=new msat_term[(to-from+1)*vars.size()];
	std::string seq,vname;
	char seqtmp[50];
	int q=0;
	for(std::set<std::string>::iterator i=vars.begin();i!=vars.end();++i) {
		for(int j=from;j<=to;j++) {
			snprintf(seqtmp,50,"%i", j); seq=seqtmp;
			vname=*i+seq;
			list[q]=msat_make_constant(env, msat_find_decl(env,vname.c_str()));
			q++;
		}
	}
	return list;
}


void prepare_apply_trans(msat_env env,product_automaton lhs, automaton rhs) {
// this function prepare global structure apply_trans_settings for apply trans
// it is called once at the boginning and the result is reused
	std::set<std::string> vars=str_set_union(lhs.vars,rhs.vars);
	apply_trans_settings.qvars=get_variables_to_quantify_out(env, 0, 0, vars);
	apply_trans_settings.size=vars.size(); //size of qvars
	for(std::set<std::string>::iterator i=vars.begin();i!=vars.end();++i){ 
		// prepare variable remapping for apply_trans 
		apply_trans_settings.var_map[*i + "1"]=*i + "0";
	}
	apply_trans_settings.predicates.push_back(msat_make_false(env)); //false is the initial predicate
}

int check_subsumption(msat_env env, product_state *x1, product_state *x2){
	// x1 >= x2 (there is a bigger chance to find an counterexample from x1)
	/* First check whether the lhs is equal */
	if (x1->l!=x2->l)
		return 0;
	/* second check whether the x1.r \subseteq x2.r --- i.e. each state in x1.r must be included in x2.r */
	for(std::set<int>::iterator i=x1->r.begin();i!=x1->r.end();++i) {
		if(x2->r.find(*i)==x2->r.end()) return 0;
	}
	/* Third check x2->abs -> x1->abs */
	return check_term_implication(env,x2->abs,x1->abs,0);
}
	
int check_subsumption_sim(msat_env env, product_state *x1, product_state *x2,simulations sim){
	/* Check whether x2 < x1 -- it x2 is subsumed by x1
	 * USE simulations
	 */
	assert(x1->l.size()==x2->l.size());
	/* first check, whether the lhs is equal or simulated */
	for(int i=0; i<x1->l.size();i++) {
		if (x1->l[i]!=x2->l[i]) {
			if(!check_term_implication(env,x2->abs,sim.lhs[i].sim[x2->l[i]-sim.lhs[i].cor][x1->l[i]-sim.lhs[i].cor],0))
				return 0;
		}
	}
	/* second, check whether each state is x1.r is equal to or simulated by some state in x2.r */
	msat_term tmp_form;
	for(std::set<int>::iterator i=x1->r.begin();i!=x1->r.end();++i) {
		int find=0;
		for(std::set<int>::iterator j=x2->r.begin();j!=x2->r.end();++j) {
			if ((*i==*j)||(check_term_implication(env,x1->abs,sim.rhs.sim[(*i)-sim.rhs.cor][(*j)-sim.rhs.cor],0))) {
				 find=1; 
				 break; 
			}
		}
		if (!find) return 0;
	}
	/* Third check x2->abs -> x1->abs */
	return check_term_implication(env,x2->abs,x1->abs,0);


}

msat_term apply_trans(msat_env env, msat_term from, msat_term trans) {
	// apply transition	
	msat_term tmp=msat_make_and(env,from,trans);
	// quantify out all variables "var0" -- only "var1" left
	tmp=msat_exist_elim_bypass(env,tmp,apply_trans_settings.qvars, apply_trans_settings.size ,MSAT_EXIST_ELIM_ALLSMT_FM);
	// remap result "var1" --> "var0"
	return term_rename(env,tmp,apply_trans_settings.var_map);
	
	
}

msat_term apply_trans_pred_abstraction(msat_env env, msat_term from, msat_term trans) {
	// apply transition	
	msat_term tmp=msat_make_and(env,from,trans);
	// compute predicate abstraction according to apply_trans_settings.predicates
	// the predicates are stored with variables "var1"
	msat_term res=msat_make_true(env);
	for(int i=0;i<apply_trans_settings.predicates.size();i++){
		if (check_term_implication(env, tmp,apply_trans_settings.predicates[i],0)) {
			res=msat_make_and(env,res,apply_trans_settings.predicates[i]);
			//std::cout <<"ADDING: "<< msat_term_repr(apply_trans_settings.predicates[i])<<"\n";
		}
	}
	return term_rename(env,res,apply_trans_settings.var_map);
}




void get_post_ll(product_state *s, std::string symbol, std::vector<int> next_lhs, msat_term lhs_form,std::map<int,msat_term> next_rhs, 
	std::queue<int> rhs_todo_states, std::set<int> rhs_states, 
	std::queue<product_state *> &out, msat_env env) {

	if (rhs_todo_states.empty()) {
		product_state *st;
		st= new product_state;
		st->l=next_lhs;
		st->r=rhs_states;
		st->symbol=symbol;
		st->seq=s->seq+1;
		st->prev=s; //add backward reference 
		// compute concrete formula
		msat_term tmp_form;
		tmp_form=lhs_form;
		for(std::map<int,msat_term>::iterator i=next_rhs.begin(); i!=next_rhs.end();i++) {
			int tmp=(*i).first;
			if(rhs_states.find(tmp)!=rhs_states.end()) 
				// *i is part of the RHS, add formula
				tmp_form=msat_make_and(env,tmp_form,next_rhs[tmp]);
			else
				// *i is not part of the RHS, add negation of formula
				tmp_form=msat_make_and(env,tmp_form,msat_make_not(env,next_rhs[tmp]));
		}
		st->trans=tmp_form;
		// concrete path formula
		st->concr=apply_trans(env,s->concr,tmp_form);
		// predicate abstraction
		st->abs=apply_trans_pred_abstraction(env,s->abs,tmp_form);
		// if st->abs==false then kill this node. No need to continue
		if (check_sat(env, st->abs)) 
			out.push(st);
		else
			delete st;
		return;
	}
	// something left to process in rhs_todo_states
	int st=rhs_todo_states.front();
	rhs_todo_states.pop();
	// we are going over all subsets -- call recursivelly without and with "st" in rhs_states 
	get_post_ll(s,symbol,next_lhs,lhs_form,next_rhs, rhs_todo_states, rhs_states, out, env);
	rhs_states.insert(st);
	get_post_ll(s,symbol,next_lhs,lhs_form,next_rhs, rhs_todo_states, rhs_states, out, env);

}

void get_post(product_state *s,std::string symbol, product_automaton lhs, int seq, std::vector<int> next_lhs,int next_lhs_move, 
	msat_term form , automaton rhs, 
	msat_env env,  std::queue<product_state *> &out) {
/* internal procedure called from compute_post
 * recursion is used to handle the on-the-fly product of the lhs */
	a_lhs rule_lhs;
	a_rhs rule_rhs;
	if (seq<s->l.size()) {
		if (lhs.components[seq].alphabet.find(symbol)==lhs.components[seq].alphabet.end()) {
			/* symbol is not a part of the alphabet in this component -- no move */
			next_lhs.push_back(s->l[seq]);
			msat_term aux=create_equalities(env,lhs,seq,symbol);
			aux=msat_make_and(env, form,aux );
			get_post(s,symbol,lhs,seq+1,next_lhs,next_lhs_move,aux,rhs,env,out);
			next_lhs.pop_back();

		} else {
			rule_lhs.src=s->l[seq];
			rule_lhs.lab=symbol;
			for(std::vector<a_rhs>::iterator j=lhs.components[seq].trans[rule_lhs].begin();j!=lhs.components[seq].trans[rule_lhs].end();++j){
				rule_rhs=*j;
				next_lhs.push_back(rule_rhs.dest);
				msat_term aux=msat_make_and(env, form, rule_rhs.formula);
				get_post(s,symbol,lhs,seq+1,next_lhs,1,aux,rhs,env,out);
				next_lhs.pop_back();
			}
		}
	} else {
		if (next_lhs_move==0) return; /* no transition via the symbol in the on-th-fly product automaton lhs --> SKIP */
		std::map<int,msat_term> next_rhs;
		std::queue<int> rhs_todo_states;
		/* first compute all transition to "new_state" via symbol and store their disjunction in next_rhs[new_state] */
		rule_lhs.lab=symbol;
		for(std::set<int>::iterator k=s->r.begin();k!=s->r.end();k++) {
			rule_lhs.src=*k;
			/* for each rule s->r[k] --- symbol ---> ... in the RHS automaton */
			for(std::vector<a_rhs>::iterator l=rhs.trans[rule_lhs].begin();l!=rhs.trans[rule_lhs].end();++l){
				rule_rhs=*l;
				if(next_rhs.find(rule_rhs.dest)==next_rhs.end()) {
					next_rhs[rule_rhs.dest]=rule_rhs.formula;
					rhs_todo_states.push(rule_rhs.dest);
				}
				else next_rhs[rule_rhs.dest]=msat_make_or(env,next_rhs[rule_rhs.dest],rule_rhs.formula);
			}

		}
		std::set<int> rhs_states;
		get_post_ll(s,symbol,next_lhs,form,next_rhs, rhs_todo_states, rhs_states, out, env);

	}
}




std::queue<product_state *> compute_post(product_state *s,product_automaton lhs,automaton rhs,msat_env env) {
// compute next states for the product state *s 
	std::queue<product_state *> out;
	std::vector<int> next_lhs;
	for(std::set<std::string>::iterator i=lhs.alphabet.begin();i!=lhs.alphabet.end();++i){
		get_post(s, *i, lhs, 0, next_lhs, 0, msat_make_true(env) , rhs, env, out);
	}
	return out;

}

int accept_config(msat_env env,product_state *s,int concrete,std::vector<automaton> lhs,automaton rhs) {
/* check whether the product state is accepting
 * concrete==0: check on abstract level
 * concrete==1: check on concrete level */
	assert(lhs.size()==s->l.size()); /* we work with a on-the-fly product on lhs */
	for(int i=0;i<lhs.size();i++) 
		if (lhs[i].final_states.find(s->l[i])==lhs[i].final_states.end()) return 0;
	int find_final=0;
	for(std::set<int>::iterator j=s->r.begin();j!=s->r.end();++j) {
		if (rhs.final_states.find(*j)!=rhs.final_states.end()) return 0;
		
	}
	if (concrete) return check_sat(env, s->concr);
	return check_sat(env,s->abs);

}

void create_init_configurations(msat_env env,std::vector<automaton> lhs, int seq, std::vector<int> lhs_state ,automaton rhs, std::vector<product_state *> &next) {
/* create initial configurations and store them into "next",
 * the recursion goes via all components of the on-the-fly product on the lhs */
	std::set<int>::iterator it;
	if (seq<lhs.size()) {
		for(it=lhs[seq].init_states.begin(); it!=lhs[seq].init_states.end();++it) {
			lhs_state.push_back(*it);
			create_init_configurations(env,lhs,seq+1,lhs_state,rhs,next);
			lhs_state.pop_back();
		}
	} else {
		product_state *st;
		st=new product_state;
		st->l=lhs_state;
		st->abs=msat_make_true(env); // true is the initial concrete and abstract state
		st->concr=msat_make_true(env);
		st->trans=msat_make_false(env);
		st->seq=0;
		st->r=rhs.init_states;
		st->prev=NULL;
		next.push_back(st);
	}
	
}


void local_ref_remove_subtree(product_state *pivot, std::vector<product_state *> &next,std::vector<product_state *> &visited) {
// remove all successors of pivot from visited/next 
// and move all states subsumed by the removed states to "next"
	std::vector<product_state *> to_remove;
	std::queue<product_state *> to_process;
	for(int i=0;i<pivot->next.size();i++) {
		to_remove.push_back(pivot->next[i]);
		to_process.push(pivot->next[i]);
	}
	while (!to_process.empty()) {
		product_state *s=to_process.front();
		to_process.pop();
		for(int i=0;i<s->next.size();i++) {
			to_remove.push_back(s->next[i]);
			to_process.push(s->next[i]);
		}
	}
	// if some of the nodes in to_remove are subsumed, remove the link them from the bigger node
	for(int i=0;i< to_remove.size();i++){
		if (to_remove[i]->subsumed_by!=NULL) {
			int j=0;
			while(1) {
				assert(j<(to_remove[i]->subsumed_by)->subsume.size());
				if ((to_remove[i]->subsumed_by)->subsume[j]==to_remove[i]) break;
				j++;
			}
			(to_remove[i]->subsumed_by)->subsume.erase((to_remove[i]->subsumed_by)->subsume.begin()+j);
			to_remove[i]->subsumed_by=NULL;
			
		}
	}
	// all nodes subsumed by a node in to_remove must be added to "next"
	for(int i=0;i< to_remove.size();i++){
		for(int j=0;j<to_remove[i]->subsume.size();j++) {
			next.push_back(to_remove[i]->subsume[j]);
			(to_remove[i]->subsume[j])->subsumed_by=NULL;
		}
	}
	//remove all nodes in to_remove from "visited" and "next"
	for(int i=0;i< to_remove.size();i++){
		int rem=0;
		while(rem<visited.size()) {
			if(to_remove[i]==visited[rem]) break;
			rem++;
		}
		if (rem<visited.size()) {
			visited.erase(visited.begin()+rem);
		} else {
			rem=0;
			while(rem<next.size()) {
				if(to_remove[i]==next[rem]) break;
				rem++;
			}
			if (rem<next.size()) {
				next.erase(next.begin()+rem);
			}
		}
	}
	for(int i=0;i< to_remove.size();i++){
		delete to_remove[i];
	}
	// remove pivot from visited and add it to next
	int rem=0;
	while(rem<visited.size()) {
		if(pivot==visited[rem]) break;
		rem++;
	}
	assert(rem<visited.size()); // pivot must be in visited
	visited.erase(visited.begin()+rem);
	pivot->next.clear(); // remove references to next elements
	next.push_back(pivot); // add pivot to TODO list
}

int dwa_inclusion(product_automaton lhs,automaton rhs,msat_env env, int bmc_limit, simulations sim,int subsume_type){
	std::cout << "\n*********************\nStarting/Restarting DWA INCLUSION\n";
	std::vector<product_state *> visited, next;
	int bmc=0;
	/* initialize next */
	std::vector<int> tmp_unused_here;
	create_init_configurations(env,lhs.components, 0, tmp_unused_here,rhs, next);
	product_state *s,*tmp;
	msat_term tmp_form;
	while ( next.size()!=0){
		s=next[0];
		next.erase(next.begin());
		std::cout << "-------------------\nProcessing path: ";
		print_path_from_root(s);
		std::cout <<"\n";
		if (DWA_VERBOSE) {
			std::cout << "concrete: " << msat_term_repr(s->concr) << "\n"
				<< "abstract: " << msat_term_repr(s->abs) << "\n";
		}
		/* do a bounded model checking */
		if (s->seq==bmc_limit) {
			std::cout << "PATH SKIPED (BMC)\n";
			bmc++;
			continue;
		}
		if (accept_config(env,s,0,lhs.components,rhs)) {
			std::cout << "ABSTRACT ACCEPT PATH\n";
			// Check whether the CE is a real one
			if (accept_config(env,s,1,lhs.components,rhs)) {
				std::cout << "RESULT: Inclusion does not hold\nError trace: ";
				print_path_from_root(s);
				std::cout << "\n";
				if(DWA_PRINT_CE) print_model(env);
				return 0;
			}
			if (DWA_LOCAL_REF) {
				// LOCAL REFINEMENT
				// find pivot 
				int pivot=s->seq-2;
				std::set<std::string> vars=str_set_union(lhs.vars,rhs.vars);
				assert(pivot>=0);
				while (!compute_interpolants(env,  s, pivot,vars)) {
					pivot--;
				}
				// find the product state with seq==pivot
				product_state *tmp=s->prev;
				while(tmp->seq!=pivot) {
					assert(tmp->seq>pivot);
					tmp=tmp->prev;
				}
				std::cout << "PIVOT: "<<*tmp<<"\n";
				std::cout << "Local refinement - removing subtree of pivot ...\n";
				local_ref_remove_subtree(tmp, next,visited);

			} else { 
				// No local refinement. Refine the abstraction and do complete restart.
				/* compute interpolants  and update predicate map */
				int res=compute_interpolants(env,  s, 0,str_set_union(lhs.vars,rhs.vars));
				assert(res);
				// restart the  computation with the new set of predicates
				return dwa_inclusion(lhs,rhs,env, bmc_limit,sim,subsume_type);
			}
		} else {
			
			int result=-1;
			for(int q=0;q<visited.size();q++) {
				int res;
				if (subsume_type>=10) res=check_subsumption_sim(env,visited[q], s, sim);
				else res=check_subsumption(env,visited[q],s);
				if (res) {
					result=q;
					break;
				}
			}
			if (result>=0) {
				std::cout <<"SUBSUMED by path: ";
				print_path_from_root(visited[result]);
				std::cout <<"\n";
				visited[result]->subsume.push_back(s); // store the subsumption relation
				s->subsumed_by=visited[result];
				continue;
			}
			for(int q=0;q<next.size();q++) {
				int res;
				if (subsume_type>=10) res=check_subsumption_sim(env,next[q], s, sim);
				else res=check_subsumption(env,next[q],s);
				if (res) {
					result=q;
					break;
				}
			}
			if (result>=0) {
				std::cout <<"SUBSUMED by path: ";
				print_path_from_root(next[result]);
				std::cout <<"\n";
				next[result]->subsume.push_back(s); // store the subsumption relation
				s->subsumed_by=next[result];
				continue;
			}
			// the actual product state is marked as visited
			// the SUBSUMED node can not be added to VISITED, becaurse it can cause cross subsumption
			visited.push_back(s);
			std::queue<product_state *> post=compute_post(s,lhs,rhs,env);
			while(!post.empty()) {
				tmp=post.front();
				post.pop();
				next.push_back(tmp);
				s->next.push_back(tmp);
				print_path_from_root(tmp);
				std::cout <<" ADDED\n";
			}
		}

	}
	if (bmc) {
		std::cout << "NO ERROR TRACE found (BMC)\nKilled paths: "<<bmc <<"\n";
	} else
		std::cout << "RESULT: Inclusion holds\n";
	return 1;

}

int main(int argc, char* argv[]){
	int bmc_limit=-1;
	int print_time=0;
	int subsume_type=0;	
	int sim_iter=-1;
	int sim_form_size=-1;
	for(int i=1; i<argc;i++) {
		if(strcmp(argv[i],"--help")==0) {
			
			std::cout << "dwa_inclusion \n" 
				<< "-b limit: apply bounded model checking -- kill paths longet then \"limit\"\n"
				<<"-c: print counterexample model\n"
				<<"-v: verbose mode\n"
				<<"-t: print times of simulation computation and dwa inclusion\n"
				<<"-gref: use global refinement and restart\n"
				<<"-lref: (default) use local refinement instead of complete restart\n";
			exit(0);
		}
		if(strcmp(argv[i],"-b")==0) {
			if (i+1>=argc) {
				std::cout << "expected: -b limit\n";
				exit(1);
			}
			bmc_limit=atoi(argv[i+1]);
			i++;
			std::cout << "Enabled BMC limit on the lenghth of the paths as: "<<bmc_limit<<"\n";
		}
		if(strcmp(argv[i],"-s")==0) {
			if (i+2>=argc) {
				std::cout << "expected: -s iteration size\n";
				exit(1);
			}
			sim_iter=atoi(argv[i+1]);
			sim_form_size=atoi(argv[i+2]);
			i=i+2;
			subsume_type=10;
		}
		if(strcmp(argv[i],"-S")==0) {
			if (i+2>=argc) {
				std::cout << "expected: -s iteration size\n";
				exit(1);
			}
			sim_iter=atoi(argv[i+1]);
			sim_form_size=atoi(argv[i+2]);
			i=i+2;
			subsume_type=-1;
		}
/*		if(strcmp(argv[i],"-x")==0) {
			if (i+2>=argc) {
				std::cout << "expected: -s iteration size\n";
				exit(1);
			}
			sim_iter=atoi(argv[i+1]);
			sim_form_size=atoi(argv[i+2]);
			i=i+2;
			subsume_type=20;
		}
		if(strcmp(argv[i],"-X")==0) {
			if (i+2>=argc) {
				std::cout << "expected: -s iteration size\n";
				exit(1);
			}
			sim_iter=atoi(argv[i+1]);
			sim_form_size=atoi(argv[i+2]);
			i=i+2;
			subsume_type=-20;
		}*/
		if(strcmp(argv[i],"-v")==0) {
			DWA_VERBOSE=1;
			std::cout << "Verbose mode on\n";
		}
		if(strcmp(argv[i],"-c")==0) {
			DWA_PRINT_CE=1;
		}
		if(strcmp(argv[i],"-t")==0) {
			print_time=1;
		}
		if(strcmp(argv[i],"-lref")==0) {
			DWA_LOCAL_REF=1;
		}
		if(strcmp(argv[i],"-gref")==0) {
			DWA_LOCAL_REF=0;
		}
	}
	product_automaton lhs;
	automaton rhs;
	msat_env env;
	msat_config cfg;	
	//int time_sim,time_inclusion;
	int time_inclusion,time_sim;
	clock_t tmp_time;
	/* create mathsat enviroment */
	cfg = msat_create_config();
	msat_set_option(cfg, "interpolation", "true");
	if (DWA_PRINT_CE) msat_set_option(cfg, "model_generation", "true");
	env = msat_create_env(cfg);
	msat_destroy_config(cfg);
	// DWA_VARIABLE_TYPE is used within creation of new variables. It must be consistent with the variables used in automata.
	// default type of variables is integer. If you like some other one, redefine it in the function get_XXX_automaton_concrete
	DWA_VARIABLE_TYPE=msat_get_integer_type(env);
	/* create automata */
	lhs=get_lhs_automaton(env);
	rhs=get_rhs_automaton(env);
	simulations sim;
	tmp_time=clock();
	if (sim_iter>0) {
		/*if ((subsume_type==20)||(subsume_type==-20))
			sim=compute_cross_simulation(env,lhs,rhs,sim_iter,sim_form_size);
		else */
		sim=compute_simulation(env,lhs,rhs,sim_iter,sim_form_size);
		if(subsume_type<0) {
			std::cout << "Finished\n";
			exit(1);
		}
	}
	time_sim=(clock()-tmp_time) / CLOCKS_PER_SEC;
	tmp_time=clock();
	prepare_apply_trans(env,lhs,rhs); // prepare global structure appply_trans_setting
	dwa_inclusion(lhs,rhs,env,bmc_limit,sim,subsume_type);
	time_inclusion=(clock()-tmp_time)/CLOCKS_PER_SEC ;
	// Print times in seconds (if requested)
	if (print_time) {
		std::cout << "---------------\nELAPTED TIMES:\n";
		if (sim_iter>0) {
			for(int i=0;i<sim.lhs.size();i++){
				std::cout <<"Simulation - LHS component no. "<<i<<": "<<  sim.lhs[i].time<<"s\n";
			}
			std::cout << "Simulation RHS: "<< sim.rhs.time<<"s\n";
			std::cout << "Simulation (total): "<< time_sim <<"s\n";

		}
		std::cout << "DWA Inclusion: "<< time_inclusion <<"s\n";
	}
	return 0;
}
